#
# FLUID test suite
#
# Copyright Texas Instruments, 2002.
# Mads Meisner-Jensen, mmj@ti.com.
#


################################################################################
# Command line constants
################################################################################

# -q  : Quiet
# -oR : No reset after download
# -y  : Redirect stderr to stdout
# -oo : Only use ROM Bootloader

$PROG_ARGS = "-oR -q -y -oo ";


################################################################################
# Constants
################################################################################

$SAY_QUIET   = 0;
$SAY_NORMAL  = 1;  # Overview of test progress
$SAY_MORE    = 2;  # More details of progress
$SAY_MUCH    = 3;  # Each test case gives details too
$SAY_DEBUG   = 4;  # Debug level detailed output
$SAY_ALL     = 5;  # Debug Debug level detailed output


################################################################################
# Globals
################################################################################

$tests_run    = 0;  # Number of testcases run
$tests_failed = 0;  # Number of testcases failed

$arg_keepgoing = 0;
$arg_dryrun   = 0;
$arg_verbose = $SAY_NORMAL;


################################################################################
# Main
################################################################################

main();

sub main()
{
    my ($time_start, $time_elapsed, $rate);

    getoptions();
    test_prog_init();
    $time_start = time();

    # FIXME: Replace 'all' with list of all test cases.
    if ($#testlist == -1) {
        push(@testlist, "all");
    }

    #test_fast_begin();

    foreach(@testlist) {
        say($SAY_NORMAL, "Running test case '$_':\n");
        if ($_ =~ /^all$/m) {
            test_args();
            test_readwrite();
            test_read();
            test_program();
        }
        elsif ($_ =~ /^bm115$/m) { test_bm115(); }
        elsif ($_ =~ /^bm921$/m) { test_bm921(); }
        else {
            say($SAY_NORMAL, "Test case unknown. Did you misspell the name?\n");
        }
    }

    $time_elapsed = time() - $time_start;
    if ($time_elapsed == 0) {
        $time_elapsed = 1;
    }
    $rate = $tests_run / $time_elapsed;

    print "\n";
    print  "==================================================\n";
    printf "$tests_run runs in $time_elapsed seconds (%.1f per second)\n", ($rate);

    if ($tests_failed == 0) {
        print "Success\n";
    }
    else {
        print "Failed $tests_failed test cases\n";
    }
}


sub getoptions
{
    foreach (@ARGV) {
	if (/^-c(.+)/) {
            $PROG_ARGS .= "$1 ";
            say($SAY_MORE,  "arg = '$1'\n");
        }
	elsif (/^-n|--dryrun/) { $arg_dryrun = 1; }
	elsif (/^-l|--list/) { listall(); }
	elsif (/^-h|--help/) { usage(); }
	elsif (/^-v|--verbose/) { $arg_verbose++; }
	elsif (/^-q|--quiet/) { $arg_verbose = -1; }
	elsif (/(\w+)/) { push(@testlist, $1); }
    }
}

sub usage
{
    print <<EOT;
aft.pl. Automatic FLUID testing.
Copyright Texas Instruments 2002. Mads Meisner-Jensen, mmj@ti.com.

Usage: [perl] aft.pl [OPTION]...

OPTIONS:
 -l         List all tests available
 -c<string> Run fluid with additional command-line arguments
 -n         Dry run - run all tests without checking results
 -v         Be verbose (repeat several times to be more verbose)
 -q         Be quiet
 -h         Display this help and exit

EOT
    exit(0);
}

sub listall
{
    print <<EOT;
Available test cases:

  all   = all available test cases
  fail  = all failing test cases
  bm115 = benchmark @ 115k2bps
  bm921 = benchmark @ 921k6bps

EOT
    exit(0);
}



################################################################################
# Output Logging
################################################################################

sub say
{
    my ($level, $str) = @_;
    my ($indent) = 0;
    if ($level < 0) {
        $level = -$level;
    }
    else {
        $indent = ($level || 1) - 1;
    }
    if ($arg_verbose >= $level) {
        print "  " x $indent . $str;
    }
}

# The logging levels are:
# 0 = Silent, only errors
# 1 = Print only test results
# 2 = Print also test info
# 3 = Print detailed test info


################################################################################
# Test Framework
################################################################################

# Run program with specified arguments
sub prog
{
    my ($args) = @_;
    $output = "";

    $prog = "fluid $PROG_ARGS $args";
    say($SAY_MORE, "Exec: '$prog' ");

    open PROG, "$prog|" or die "'$prog' failed";

    while (<PROG>) {
        chop;
        if ($_ ne "") {
            # Discard white space at both ends of line
            $_ =~ s/^\s*(.*?)\s*$/$1/;
            $output .= $_ . " ";
        }
    }
    say(-$SAY_MORE, "ok\n");
    say($SAY_MUCH,  "Out : '$output'\n");

    $tests_run++;

    return $output;
}

# Expect and test versus an exact output
sub prog_exp
{
    ($command, $expect) = @_;

    $output = prog($command);
    $arg_dryrun || expect($expect);
}

# Expect error
sub prog_exp_error
{
    ($command, $expect) = @_;

    $output = prog($command);
    $arg_dryrun || expect_regexp("ERROR\\((\\d+)\\)", "$expect");
}

# Expect and test versus a regular expression output
sub prog_exp_regexp
{
    ($command, $regexp, $expect) = @_;

    $output = prog($command);
    $arg_dryrun || expect_regexp($regexp, $expect);
}

# Test for specific value
sub expect
{
    my ($expect) = @_;

    # FIXME: We should coalesce all multiple-space occurrences with
    # single-spaces.

    # Discard white space at both ends of line
    ($newoutput = $output) =~ s/^\s*(.*?)\s*$/$1/;
    ($newexpect = $expect) =~ s/^\s*(.*?)\s*$/$1/;

    if ($newexpect ne $newoutput) {
        print "TEST-FAIL:\n" .
            "      expect('$newexpect'):\n" .
            "         got '$newoutput'\n";
        if (!$arg_keepgoing) {
            die "Testing stopped";
        }
    }
}

sub expect_regexp
{
    my ($regexp, $expect) = @_;

    $output =~ m/$regexp/;
    $match = $1;

    if ($expect ne $match) {
        print "TEST-FAIL:\n" .
              "  expect_regexp('$regexp', '$expect'):\n" .
              "            got '$output'\n";
        $arg_keepgoing || die "Testing stopped";
    }
}


################################################################################
# Test cases
################################################################################

sub test_prog_init
{
    say($SAY_NORMAL, "Automatic Fluid Testing (for Calypso only)\n");
    say($SAY_NORMAL, "If test freezes, ensure internal bootloader is enabled\n");
    say($SAY_NORMAL, "======================================================\n");

    open PROG, "fluid -V|" or die "'fluid -V' failed";
    $tmp = <PROG>;
    ($prog_rev) = ($tmp =~ /[A-Za-z ]+([0-9]\.[0-9]+).*/m);

    say($SAY_NORMAL, "Fluid revision = $prog_rev\n");

    open PROG, "fluid -dv|" or die "'fluid -dv' failed";
    say($SAY_NORMAL, <PROG>);

    say($SAY_NORMAL, "======================================================\n");
}

sub test_fast_begin
{
    # Start target command interpreter, so we save this step in all future
    # tests

    prog("-n -f amd.m0");
    $PROG_ARGS .= "-dc ";

    say($SAY_NORMAL, "Using command-line: fluid $PROG_ARGS\n");
    say($SAY_NORMAL, "================================================\n");
}


################################################################################
# Test cases
################################################################################

sub test_args
{
    say($SAY_NORMAL, "Testing: args\n");

    prog_exp_error("-q -p9", 5);
}

sub test_readwrite
{
    say($SAY_NORMAL, "Testing: readwrite\n");

    prog("-w80k..81k=1,2:65536=3:66k=\"\"\"\"foo bar\"\"\"\",10,11,12");
    prog_exp("-sx -r81920..81922",
             "14000 ..               01 02");
    prog_exp("-sx -r0x143fe..0x14402",
             "143fe ....             01 02 ff ff");
    prog_exp("-sx -r64k..65538",
             "10000 ..               03 ff");
    prog_exp("-sx -r66k..0x10840",
             "10800 foo bar......... 66 6f 6f 20 62 61 72 0a 0b 0c ff ff ff ff ff ff " .
             "10810 ................ ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff " . 
             "10820 ................ ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff " .
             "10830 ................ ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ");
}

sub test_read
{
    say($SAY_NORMAL, "Testing: read\n");

    # Negative tests

    prog_exp_error("-r2M..", 70);   # Bad address range
    prog_exp_error("-r0", 70);
    prog_exp_error("-r1k..2k", 59); # No output file

    # Positive tests

    prog_exp("-r0x300000..0x300100 -f test.m0", "");
    # FIXME: re-program m0 file, then read it back and compare.
}

sub test_program
{
    say($SAY_NORMAL, "Testing: program\n");

    # Negative tests

    prog_exp_error("-fnofile", 50);   # File open error
}


################################################################################
# Benchmarking
################################################################################

sub test_bm_read
{
    my ($baudrate) = @_;
    $kb = $bm_size / 1024;

    if ( ! -f $bm_file ) {
        say($SAY_MORE, "  Reading ${kb}kB into '$bm_file'\n");
        prog_exp("-r0..$bm_size -f $bm_file -b $baudrate", "");
    }
    else {
        say($SAY_MORE, "  Reusing already existing '$bm_file'\n");
    }
}

sub test_bm
{
    my ($baudrate) = @_;

    $bm_size = 1024 * 1024;
    $kb = $bm_size / 1024;
    $bm_file = "benchmark_$kb.m0";

    say($SAY_NORMAL, "Benchmarking @ ${baudrate}kbps\n");

    test_bm_read($baudrate);

    $output = prog("-f $bm_file -C -b$baudrate -v -v");
    if ($output =~ m/ERROR/) {
        print "ERROR launching fluid: Bad baudrate?\n";
        $s = 1000;
    }
    else {
        $output =~ m/Used time: \([^)]*\)\s+(\d+)s/;
        $s = $1;
    }

    $kbps = $kb / $s;
    $spmb = 1024 / $kbps;

    print  "==================================================\n";
    printf "Program speed: ${kb}kB in ${s}s = %.1fkB/s = %.0fs/MB\n", $kbps, $spmb;
    print  "==================================================\n";
}

sub test_bm115
{
    test_bm(115);
}

sub test_bm921
{
    test_bm(921);
}

