/******************************************************************************
 * FLUID (Flash Loader Utility Independent of Device)
 *
 * (C) Delta Technologies 2001.
 * Cleanup, modifications and extensions by Mads Meisner-Jensen, mmj@ti.com.
 *
 * Command Interpreter State Machine
 *
 * $Id: cmd.c 1.18.1.26 Mon, 28 Apr 2003 08:49:16 +0200 tsj $
 *
 ******************************************************************************/

#include "target.h"
#include "../pc/lz.h"
// IMEI Protection
#include "../inc/secure_types.h"

#define C_IMEISV_BYTES 8
#define NULL (UWORD32 *) 0x00000000L
// End IMEI Protection

#define NEWZIP 0

#define DEBUG_TRACE 1

#if (DEBUG_TRACE == 1)
#define DEBUG_ERROR         0
#define DEBUG_LED           1
#define DEBUG_LED_PROGRESS  1
#define DEBUG_TRACE_INIT    0
#define DEBUG_TRACE_LZ      0
#else
#define DEBUG_ERROR         0
#define DEBUG_LED           0
#define DEBUG_LED_PROGRESS  0
#define DEBUG_TRACE_INIT    0
#define DEBUG_TRACE_LZ      0
#endif


// NOTE that in this module *all* variables are local. There are no global
// nor static variables. This is important because otherwise the module is
// not completely independent of where the RAM memory is! This way, we can
// use the same cmd.m0 file for both Ulysses and Calypso devices.


/******************************************************************************
 * Globals
 ******************************************************************************/

// The bss data is located at an offset after the program code,
#define BSS_OFFSET (8 * 1024)

// Address assigned when no erase is in progress
#define NO_ERASE 0xFFFFFFFF


/******************************************************************************
 * Prototypes
 ******************************************************************************/

int    ram_size(bss_t *bss, volatile char *addr);
int    hardware_init(bss_t *bss);
int    hardware_init_ulysses(bss_t *bss);
int    hardware_init_calypso(bss_t *bss);
int    dpll_lock(bss_t *bss);
void   uart_init(bss_t *bss, UBYTE dlh, UBYTE dll);
void   loop(bss_t *bss, char ch);

void   putchar(bss_t *bss, uint8 ch);
void   putchar_wait(bss_t *bss);
void   put32(bss_t *bss, unsigned int value);
uint8  getchar(bss_t *bss);
uint8  getchar_fast(bss_t *bss);
uint32 getword(bss_t *bss);
uint8  getchar_is_ready(bss_t *bss);
void   delay(int ms);

void timer_start(bss_t *bss);
int  timer_read(void);
int  timer_elapsed_since(bss_t *bss, int last);
int  timer_elapsed(bss_t *bss);

void  flash_detect(bss_t *bss);
void  flash_query_cfi_data(bss_t *bss, uint16 *data, uint16 offset);
int   recv(bss_t *bss);
int   lzdec(bss_t *bss);
int   flash_machine(bss_t *bss);
void  flash_machine_reset(bss_t *bss);
void  led_progress(bss_t *bss, int addr);

// Prototypes for functions in debug.c
void debug_putstr(bss_t *bss, char *string);
void debug_putnum(bss_t *bss, unsigned int number);

int  fifo_is_empty(bss_t *bss);
void fifo_reset(bss_t *bss);
int  fifo_is_full(bss_t *bss);
struct block_s *fifo_pop(bss_t *bss);
void fifo_push(bss_t *bss);
uint32 fifo_peek(bss_t *bss);


#if (DEBUG_LED == 1)
void led_set(bss_t *bss, uint16 n);
void led_clr(bss_t *bss, uint16 n);
void led_on(bss_t *bss, uint16 n);
void led_off(bss_t *bss, uint16 n);
void led_toggle(bss_t *bss, uint16 n);
#else
#define led_set(foo, bar)
#define led_clr(foo, bar)
#define led_on(foo, bar)
#define led_off(foo, bar)
#define led_toggle(foo, bar)
#endif

// NOTE: When changing any of these, remember to update the help text in
// fluid.c:main_debug_usage()
enum {
    LED_RECV     =  0,
    LED_ERASE    =  1,
    LED_PROGRAM  =  2,
    LED_OVERHEAD =  3,
    LED_IDLE     =  4,

    LED_BUG      =  5,

    LED_TOGGLE   =  6,
    LED_BUSY     =  7,

    LED_RAMCS0   = 15
};

 /******************************************************************************
 * Functions
 ******************************************************************************/

// Note that we enter this code in 32-bit mode so we immediately change to
// 16-bit (thumb) mode.
asm("        .global _c_int00");
asm("        .state32");
asm("_c_int00: ");

// Get our execution address and set the stack pointer to just below this
// code. We assume we have been loaded at an address a multiple of 0x1000
// bytes.
asm("        mov     r0, pc");
asm("        bic     r0, r0, #0x0FF");
asm("        bic     r0, r0, #0xF00");
asm("        mov     r1, r0");
asm("        mov     sp, r0");

asm("        adr     lr,start_addr + 1 ; address with lsb set");
asm("        bx      lr                ; branch and shift to thumb mode");
asm("        .state16");
asm("start_addr:");

void main(int pc, int sp)
{
    bss_t *bss;
    char ch, *p;

    bss = (bss_t *) (pc + BSS_OFFSET);

    bss->pc = pc;
    bss->compress = 0;

    hardware_init(bss);

    bss->uart = UART_MODEM;
    uart_init(bss, 0, 0x07); // baudrate = 115200 bps
    bss->uart = UART_IRDA;
    uart_init(bss, 0, 0x07); // baudrate = 115200 bps

    // Detect which UART we run on...
    while (1) {
        bss->uart = UART_MODEM;
        if ((REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY))
            break;
        bss->uart = UART_IRDA;
        if ((REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY))
            break;
    }

#if (DEBUG_TRACE_INIT == 1)
    debug_putstr(bss, "PC, SP, bss:");
    debug_putnum(bss, pc);
    debug_putnum(bss, sp);
    debug_putnum(bss, (uint32) bss);
#endif

    ROM_WRITE_ENABLE();

    fifo_reset(bss);
    flash_machine_reset(bss);
    timer_start(bss);
    bss->time.start = timer_read();

    led_clr(bss, 0xFF);

    for (;;)
    {
        union16_t size16, data16;

        ch = getchar(bss);

        if (ch != PROTO_HELLO)
          continue;

        putchar(bss, PROTO_READY);
        data16.i = bss->chip_id;
        putchar(bss, data16.b[0]);
        putchar(bss, data16.b[1]);
        putchar(bss, (bss->ramcs0  ? 'R' : ' ' ));
        ch = ram_size(bss, (char *) &bss->buffer[0]);
        putchar(bss, ch);
        putchar(bss, bss->fifo.max);
        break;
    }

    for (;;)
    {
        ch = getchar(bss);
        loop(bss, ch);
    }
}

void loop(bss_t *bss, char ch)
{
    int i, n, length;
    union16_t size16, data16;
    uint32 size32, cksum32, addr;
    uint16 *dst, *src;
    uint8 cksum8;
    char *p;
    char xxo;
    // IMEI Protection
    int             d_error;
    UWORD32         a_imeisv[C_IMEISV_BYTES / sizeof(UWORD32)];
    E_SECURE_STATUS d_status;
    // End IMEI Protection

    led_clr(bss, 0xff);
    led_set(bss, ch);

    switch (ch) {

    case PROTO_BAUDRATE:
        xxo = (getchar(bss) == PROTO_BAUDRATE_XXO); // Get XXO flag
        ch = getchar(bss); // Get DLH
        // Set divider register
        // Apparently, we cannot read from the UART_LCR register!?
        //*(volatile uint8 *) (bss->uart + UART_LCR) |= 0x80;
        *(volatile uint8 *) (bss->uart + UART_LCR) = 0x80;
        *(volatile uint8 *) (bss->uart + UART_RHR) = ch;
        *(volatile uint8 *) (bss->uart + UART_LCR) = 0x03; // 8,n,1
        //*(volatile uint8 *) (bss->uart + UART_LCR) &= ~0x80;

        {
            // A small delay is required here.
            // Otherwise, we reply before the PC has changed baud rate...
            int i;
            for (i = 0; i < 100000; i++)
                asm (" NOP ");
        }

        ch = PROTO_READY;

        // If xxo is set, we should enable the eXternal Xtal Oscillator
        // (currently only DSample with a 14MHz oscillator)
        if (xxo) {
            // DPLL bypass mode
            *(volatile uint16 *) MEM_DPLL_ADDR &= 0xffef;
            // Start 14MHz oscillator
            *(volatile uint16 *) DS_CTRL = DS_CTRL_CLK14_ON;
            delay(10);
            // Switch to 14MHz clock
            *(volatile uint16 *) DS_CTRL =
                DS_CTRL_CLK14_ON | DS_CTRL_CLK14_ENA;

            // Set to "13MHz" mode (div 1) CALPLUS JP
            *(volatile unsigned char*) 0xfffffd02 = 0x11;

            // Re-activate DPLL and wait for lock
            *(volatile uint16 *) MEM_DPLL_ADDR |= 0x0010;
            if ((ch = dpll_lock(bss)) == 0)
                ch = PROTO_READY;
        }
        putchar(bss, ch);
        break;

    case PROTO_INIT:
        ch = getchar(bss); // Get target type
        switch (ch) {
        case 'C':
        case 'P':
        case 'c': ch = hardware_init_calypso(bss); break;
        case 'U': ch = hardware_init_ulysses(bss); break;
        default:  ch = PROTO_ERROR; break;
        }

        if (ch == 0)
            ch = PROTO_READY;
        putchar(bss, ch);
        break;

    case PROTO_DETECT:
        flash_detect(bss);
        p = (char *) &bss->buffer[0];
        for (i = 0; i < 12; i++)
            putchar(bss, p[i]);

        // Query flash write buffer size. This information is needed by the
        // intel buffer write driver (intelbw).
        flash_query_cfi_data(bss, &bss->flash_buf_size, 0x2A<<1);
        bss->flash_buf_size = 1 << bss->flash_buf_size; // Read as 2^n
        break;

    case PROTO_ERASE:
        // Note that one cannot save time by programming multiple
        // sectors in one go. Not on current flash devices anyway.
        timer_start(bss); // start the timing process
        size16.b[0] = getchar(bss);
        size16.b[1] = getchar(bss);
        bss->erase.list_size = size16.i;
        bss->erase.index = 0;
        if (bss->erase.list_size < 0 ||
            ERASE_LIST_SIZE_MAX < bss->erase.list_size) {
            putchar(bss, PROTO_ERROR_MEMORY);
            break;
        }
        for (i = 0; i < size16.i; i++) {
            bss->erase.list[i] = getword(bss);

            if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
                (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A))
                bss->erase.list[i] += CALP_OFFSET;

        }
        putchar(bss, PROTO_READY);

        if (bss->ramcs0) {
            // We don't erase the RAM --- that would be silly!
            bss->erase.addr = NO_ERASE;
            while (bss->erase.list_size--)
                putchar(bss, PROTO_ERASE);
        }
        break;

    case PROTO_FLASH_START:
        bss->compress = getchar(bss);
        putchar(bss, PROTO_READY);
        if (bss->compress)
            decompress_init(&bss->lz);
        ch = flash_machine(bss);
        flash_machine_reset(bss);
        putchar(bss, ch);
        break;

    case PROTO_IMEI_PROTECT:
        // Get the 8 bytes IMEI-SV
        a_imeisv[0]  = ((UWORD32) getchar(bss)) << 24;
        a_imeisv[0] |= ((UWORD32) getchar(bss)) << 16;
        a_imeisv[0] |= ((UWORD32) getchar(bss)) << 8;
        a_imeisv[0] |= ((UWORD32) getchar(bss));

        a_imeisv[1]  = ((UWORD32) getchar(bss)) << 24;
        a_imeisv[1] |= ((UWORD32) getchar(bss)) << 16;
        a_imeisv[1] |= ((UWORD32) getchar(bss)) << 8;
        a_imeisv[1] |= ((UWORD32) getchar(bss));
        putchar(bss, PROTO_READY);

        // Get the platform certificate address
        addr  = ((UWORD32) getchar(bss)) << 24;
        addr |= ((UWORD32) getchar(bss)) << 16;
        addr |= ((UWORD32) getchar(bss)) << 8;
        addr |= ((UWORD32) getchar(bss));
        putchar(bss, PROTO_READY);

        // Check if address range is erased
        size32 = (sizeof(T_PLATFORM_CERTIFICATE) + C_IMEISV_BYTES) / 2;
        src = (uint16 *) addr;
        d_error = 0;
        while (size32--) {
            data16.i = *src++;
            if (data16.b[0] != 0xFF || data16.b[1] != 0xFF) {
                d_error = 1;
                break;
            }
        }

        if (d_error) {
            putchar(bss, PROTO_ERROR_VERIFY);
            break;
        }
        else
            putchar(bss, PROTO_READY);

        // Prepare the platform certificate and binding struct
        bss->d_platform_certificate.d_Certsize    = 0;
        bss->d_platform_certificate.d_Certtype    = C_CERTTYPE_PLAT;
        bss->d_platform_certificate.d_Confrequest = C_CRYPTNOREQUEST;
        bss->d_platform_certificate.d_Addcode     = addr + sizeof(T_PLATFORM_CERTIFICATE);
        bss->d_platform_certificate.d_Codesize    = C_IMEISV_BYTES;
        bss->d_platform_certificate.d_AppID       = 0x00000001L;
        bss->d_platform_certificate.a_iv[0]       = 0x12345678L;
        bss->d_platform_certificate.a_iv[1]       = 0x90ABCDEFL;

        bss->d_bind.p_Cert       = (T_PLATFORM_CERTIFICATE*) &bss->d_platform_certificate;
        bss->d_bind.p_Code       = (UWORD32*) a_imeisv;
        bss->d_bind.p_CertCtx    = (T_PLATFORM_CERTIFICATE*) NULL;
        bss->d_bind.p_BindCtx    = (T_BINDCTX*) NULL;
        bss->d_bind.d_Steplength = 0L;
        bss->d_bind.d_timeout    = 0L;
        bss->d_bind.b_start      = C_TRUE;

        // Run-Time Loader (Binding) Secure ROM Service
        bss->pf_ROM_Sservice = (T_ROM_SSERVICE) C_ROM_SSERVICE_ENTRY_POINT_ADDR;

        *(volatile uint16 *) RNG_CONFIG = RNG_CONFIG_DEFAULT;
        *(volatile uint16 *) RNG_ALARM  = 0x0;

        d_status = C_SECURE_RNG_ALARM;
        while (d_status == C_SECURE_RNG_ALARM)
            d_status = bss->pf_ROM_Sservice(C_SEC_SERVICE_BINDING_ID,
                                            (T_SSERVICE_BIND *)&bss->d_bind,
                                            NULL);

        if (d_status != C_SECURE_SUCCESS)
            putchar(bss, PROTO_ERROR_ROM_SSERVICE);
        else {
            putchar(bss, PROTO_FLASH_START);

            // Program platform certificate
            dst = (uint16 *) addr;
            src = (uint16 *) &bss->d_platform_certificate;

            d_error = bss->flash->program(bss, dst, src, sizeof(T_PLATFORM_CERTIFICATE) / 2);
            if (d_error == 0)
                putchar(bss, PROTO_PROGRAM);
            else {
                putchar(bss, d_error);
                break;
            }

            // Program IMEI-SV
            dst = (uint16 *) (addr + sizeof(T_PLATFORM_CERTIFICATE));
            src = (uint16 *) a_imeisv;

            d_error = bss->flash->program(bss, dst, src, C_IMEISV_BYTES / 2);
            if (d_error == 0)
                putchar(bss, PROTO_PROGRAM);
            else {
                putchar(bss, d_error);
                break;
            }
        }
        break;

    case PROTO_CKSUM:
#if 1
        n = getchar(bss);
        size32 = getword(bss) / 2;

        // Note that we depend on the UART FIFO to buffer the bytes that we
        // receive while calculating checksums
        for (i = 0; i < n; i++) {
            led_on(bss, LED_BUSY);
            addr = getword(bss);

            if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
                (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A))
                addr += CALP_OFFSET;

            src = (uint16 *) addr;
            cksum32 = 0;
            length = size32;
            while (length--) {
                cksum32 += *src * (((uint16) src & 0xFFFF));
                src++;
            }
            led_off(bss, LED_BUSY);
            putchar(bss, cksum32 >>  0);
            putchar(bss, cksum32 >>  8);
            putchar(bss, cksum32 >> 16);
            putchar(bss, cksum32 >> 24);
        }
#else
        // This code is if we transfer all checksums at once
        size16.b[0] = getchar(bss->uart);
        size16.b[1] = getchar(bss->uart);

        p32 = bss->buffer;
        for (i = 0; i < size16.i; i++) {
            size32.b[0] = getchar(bss->uart);
            size32.b[1] = getchar(bss->uart);
            size32.b[2] = getchar(bss->uart);
            size32.b[3] = getchar(bss->uart);
            addr.b[0] = getchar(bss->uart);
            addr.b[1] = getchar(bss->uart);
            addr.b[2] = getchar(bss->uart);
            addr.b[3] = getchar(bss->uart);
            *p++ = size32.i;
            *p++ = addr.i;
        }

        p32 = bss->buffer;
        for (i = 0; i < size16.i; i++) {
            size32.i = *p++ / 2;
            mem = (uint16 *) *p++;
            bigcksum = 0;
            while (size32.i--) {
                bigcksum += *mem * (((uint16) mem & 0xFFFF));
                mem++;
            }
            putchar(bss, bigcksum >>  0);
            putchar(bss, bigcksum >>  8);
            putchar(bss, bigcksum >> 16);
            putchar(bss, bigcksum >> 24);
        }
        putchar(bss, PROTO_READY);
#endif
        break;
#if 0
    case PROTO_ISEMPTY:
        size32 = getword(bss);
        addr = getword(bss);
        src = (uint16 *) addr;
        cksum32 = 0;
        size32 = size32 / 2;
        while (size32--)
            cksum32 += (*src++ == 0xFFFF);

        if (cksum32)
            putchar(bss, '0'); // false: is *not* empty
        else
            putchar(bss, '1'); // true: *is* empty
        break;
#endif
    case PROTO_READ:
        size32 = getword(bss) / 2;
        addr = getword(bss);
        src = (uint16 *) addr;
        led_progress(bss, addr);
        cksum8 = 0;
        while (size32--) {
            data16.i = *src++;
            putchar(bss, data16.b[0]);
            putchar(bss, data16.b[1]);
            cksum8 ^= data16.b[0];
            cksum8 ^= data16.b[1];
        }
        putchar(bss, cksum8);
        break;

    case PROTO_DOWNLOAD:
        addr = getword(bss);
        size16.b[0] = getchar(bss);
        size16.b[1] = getchar(bss);
        dst = (uint16*) addr;
        bss->flash = (flash_t *) addr;
        putchar(bss, PROTO_READY);
        while (size16.i--) {
            // Data bytes arrive in reverse order due to little endian mode
            data16.b[1] = getchar(bss);
            data16.b[0] = getchar(bss);
            *dst++ = data16.i;
        }
        // Compute functions addresses from the offsets in <flash>
        ((uint32*) bss->flash)[0] += (uint32) bss->flash;
        ((uint32*) bss->flash)[1] += (uint32) bss->flash;
        ((uint32*) bss->flash)[2] += (uint32) bss->flash;

        // Secure Calypso Plus uses cmdp.cmd as linker file in which the start
        // address of the flash programmer is 0x08020000.
        if ((uint32)recv >= CMD_SECURE_CALP_OFFSET) {
            bss->flash->getchar = (uint8 (*)(void)) ((uint32) getchar);
            bss->flash->putchar = (void (*)(uint8)) ((uint32) putchar);
            bss->flash->debug_putstr = (void (*)(uint32, char *)) ((uint32) debug_putstr);
            bss->flash->debug_putnum = (void (*)(uint32, unsigned int)) ((uint32) debug_putnum);
            bss->flash->uart = bss->uart;
            bss->recv = (int (*)(struct bss_s *)) ((uint32) recv);
        }
        else {
            bss->flash->getchar =
                (uint8 (*)(void)) (bss->pc + (uint32) getchar);
            bss->flash->putchar =
                (void (*)(uint8)) (bss->pc + (uint32) putchar);
            bss->flash->debug_putstr =
                (void (*)(uint32, char *)) (bss->pc + (uint32) debug_putstr);
            bss->flash->debug_putnum =
                (void (*)(uint32, unsigned int)) (bss->pc + (uint32) debug_putnum);
            bss->flash->uart = bss->uart;
            bss->recv =
                (int (*)(struct bss_s *)) (bss->pc + (uint32) recv);
        }
        putchar(bss, bss->flash->init(addr));
        break;

    case PROTO_RESET:
        // Restart the target by jumping to address zero. Send
        // acknowledge twice... The transmission of the second char is
        // interrupted by the reset.

        putchar(bss, PROTO_READY);
        putchar_wait(bss);
        // Switch back to 13MHz clock (only on DSample)
        *(volatile uint16 *) DS_CTRL = 0;

        ROM_WRITE_DISABLE();

        // Setup watchdog timer and let it timeout to make a reset
        *(volatile uint16*) 0xfffff804 = 0xFFFF;  // Timer to watchdog
        *(volatile uint16*) 0xfffff800 = 0x0080;  // Start timer
        // Apparently works it only if we read this register?
        i = *(volatile uint16*) 0xfffff802;
        *(volatile uint16*) 0xfffff802 = 0x0001;  // Load timer

        while(1);

        break;

    case PROTO_TIMERS:
        put32(bss, bss->time.erase);
        put32(bss, bss->time.program);
        put32(bss, bss->time.recvonly);
        put32(bss, bss->time.recv);
        put32(bss, bss->time.comm);
        put32(bss, bss->time.setup);
        put32(bss, bss->time.overhead);
        put32(bss, bss->time.dezip);
        break;
    default:
        putchar(bss, PROTO_ERROR);
    }
}


/******************************************************************************
 * Flash State Machine
 ******************************************************************************/

int flash_machine(bss_t *bss)
{
    struct block_s *block;
    uint32 addr;
    uint16 *src, *dst;
    int i, error, j;

    bss->recv_state = 0;
    bss->flash_end = 0;
    error = 0;
    bss->time.setup = timer_elapsed_since(bss, bss->time.start);
    bss->time.start = timer_read();

    while (!bss->flash_end)
    {
        // First try to do a program... If FIFO is not empty, then ensure
        // that sector for this data block has been erased
        if (!fifo_is_empty(bss) && fifo_peek(bss) < bss->erase.addr)
        {
            block = fifo_pop(bss);
            dst = (uint16 *) block->addr;
            src = (uint16 *) &block->data[0];

            led_on(bss, LED_PROGRAM);
            led_progress(bss, (int) dst);

            bss->time.recvonly += timer_elapsed(bss);
            //ROM_WRITE_ENABLE();
            error = bss->flash->program(bss, dst, src, block->size / 2);
            //ROM_WRITE_DISABLE();
            bss->time.program += timer_elapsed(bss);

            // Verify data. Run recv() to avoid FIFO overflow while we verify.
            if (error == 0) {
                for (j = 0; j < 16; j++) {
                    for (i = 0; i < block->size / 2 / 16; i++) {
                        if (*dst++ != *src++)
                            error = PROTO_ERROR_VERIFY;
                    }
                    if (error)
                        break;
                    if (error = recv(bss) != 0)
                        break;
                }
            }
            if (error == 0)
                putchar(bss, PROTO_PROGRAM);
            else {
                led_on(bss, LED_BUG);
#if (DEBUG_ERROR == 1)
                debug_putstr(bss, "flash_program @ ");
                debug_putnum(bss, block->addr);
#endif
                putchar(bss, error);
                break;
            }

            led_off(bss, LED_PROGRAM);
            bss->time.overhead += timer_elapsed(bss);
        }

        // Now try to do an erase...
        else if (bss->erase.index < bss->erase.list_size)
        {
            bss->erase.addr = bss->erase.list[bss->erase.index++];
            led_on(bss, LED_ERASE);
            bss->time.recvonly += timer_elapsed(bss);
            //ROM_WRITE_ENABLE();
            error = bss->flash->erase(bss, bss->erase.addr);
            //ROM_WRITE_DISABLE();
            bss->time.erase += timer_elapsed(bss);

            if (error == 0)
                putchar(bss, PROTO_ERASE);
            else {
                // An error has occurred
                led_on(bss, LED_BUG);
#if (DEBUG_ERROR == 1)
                debug_putstr(bss, "flash_erase @ ");
                debug_putnum(bss, bss->erase.addr);
#endif
                putchar(bss, error);
                break;
            }
            // Was this the last sector?
            if (bss->erase.index >= bss->erase.list_size)
                bss->erase.addr = NO_ERASE;

            led_off(bss, LED_ERASE);
        }
        else {
            // We end up here when we can neither program nor erase. If the
            // flash_end flag is set, we are finished with all our work...
            if (bss->flash_end)
                break;
            error = bss->recv(bss);
        }
        if (error)
            return error;
    }
    return PROTO_FLASH_END;
}

int recv(bss_t *bss)
{
    char ch;
    uint16 *src, *dst;
    enum {
        IDLE,
        HEADER,
        DATA,
        TRAILER
    };

    if ((REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RX_OE)) {
        led_on(bss, LED_BUG);
        return PROTO_ERROR_FIFO_OVERFLOW;
    }

    while (getchar_is_ready(bss))
    {
        ch = getchar_fast(bss);

        switch (bss->recv_state) {
        case IDLE:
            // Waiting for next buffer...
            // FIXME: Somewhere we should have some kind of flow control so
            // the PC does not send more blocks than we have room for!
            if (ch == PROTO_PROGRAM) {
                bss->time.comm += timer_elapsed_since(bss, bss->time.start);
                bss->time.start = timer_read();
                led_off(bss, LED_IDLE);
                led_on(bss, LED_RECV);
                led_toggle(bss, LED_TOGGLE);
                bss->rx.index = 0;
                bss->recv_state = HEADER;
            }
            else if (ch == PROTO_FLASH_END) {
                bss->flash_end = 1;
            }
            else {
                return PROTO_ERROR; // Protocol error!
            }
            break;

        case HEADER:
            // Collect address and size of block
            bss->rx.buf[bss->rx.index++] = ch;
            if (bss->rx.index >= 8) {
                bss->rx.size  = ((struct block_s *) &bss->rx.buf)->size;
                bss->rx.addr  = ((struct block_s *) &bss->rx.buf)->addr;

                if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
                    (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A))
                    bss->rx.addr += CALP_OFFSET;

                bss->rx.rem   = bss->rx.size;
                // If we are in compress mode, we load data into the top of the
                // buffer. Otherwise we just load into the start of the buffer.
                bss->rx.index = (bss->compress ?
                                 sizeof(bss->rx.buf) - bss->rx.size: 0);
                bss->rx.cksum = 0;
                // Check that size is sensible
                if (bss->rx.size > sizeof(bss->rx.buf)) {
#if (DEBUG_ERROR == 1)
                    debug_putstr(bss, "block_size=");
                    debug_putnum(bss, bss->rx.size);
#endif
                    return PROTO_ERROR_MEMORY;
                }
                // putchar(bss, PROTO_READY); // FIXME: to be optimized out
                bss->recv_state = DATA;
            }
            break;

        case DATA:
            // Collect block data and dezip them
#if (NEWZIP == 0)
            bss->rx.buf[bss->rx.index++] = ch;
            bss->rx.cksum += (uint8) ch;
            bss->rx.rem--;
#else
            lzdecode(ch);
#endif
            if (bss->rx.rem == 0)
                bss->recv_state = TRAILER;
            break;

        case TRAILER:
            bss->rx.cksum += (uint8) ch;
            if (bss->rx.cksum != 0)
                return PROTO_ERROR_CKSUM;

            led_off(bss, LED_RECV);
            bss->time.recv += timer_elapsed_since(bss, bss->time.start);

#if (NEWZIP == 0)
            led_on(bss, LED_OVERHEAD);
            if (bss->compress) {
                bss->time.start = timer_read();
                bss->rx.size =
                    decompress(&bss->lz,
                               &bss->rx.buf[0],
                               &bss->rx.buf[sizeof(bss->rx.buf) - bss->rx.size],
                               bss->rx.size);
                bss->time.dezip += timer_elapsed_since(bss, bss->time.start);
            }
            if (bss->ramcs0) {
                // We are "programming" the RAM so copy block to RAM
                dst = (uint16 *) bss->rx.addr;
                src = (uint16 *) &bss->rx.buf[0];
                bss->rx.size = bss->rx.size / 2;
                while (bss->rx.size--)
                    *dst++ = *src++;
            }
            else {
                // Store the received block into the FIFO
                fifo_push(bss);
            }
            led_off(bss, LED_OVERHEAD);
#endif
            bss->time.start = timer_read();
            led_on(bss, LED_IDLE);

            putchar(bss, PROTO_READY);
            bss->recv_state = IDLE;
            break;
        }
    }

    return 0;
}

void lzdecode(bss_t *bss, char ch)
{
#if (NEWZIP == 1)
    switch (bss->lzdec_state) {
    case MASK:
        bss->mask = ch;
        bss->masksize = 8;
        bss->lzdec_state = (mask & 1 ? Z1 : BYTE);
        break;
    case BYTE:
        history[hist_pos++] = ch;
        *buf++ = ch;
        bss->lzdec_state = (mask & 1 ? Z1 : BYTE);
        break;
    case Z1:
        bss->zword[0] = ch;
        bss->lzdec_state = Z2;
        bss->mask <<= 1;
        bss->masksize++;
        break;
    case Z2:
        bss->zword[1] = ch;
        pos = zword &  posmask;
        len = zword >> lenbits;
        memcpy(buf, &history[pos], len);
        buf += len;
        lzdec_state = (mask & 1 ? Z1 : BYTE);
        break;
    }
    bss->mask >>= 1;
    bss->masksize--;
    if (bss->masksize == 0)
        bss->lzdec_state = MASK;
#endif
}


/******************************************************************************
 * Initialization and Utility
 ******************************************************************************/

// Assembly code that satisfies compiler's need of an indirect call
asm("IND$CALL:");
asm("        MOV     R12, R4");
asm("        LSR     R4, R4, #0x1");
asm("        BCC     ep_32");
asm("        BX      R12");
asm("        MOV     R8, R8");
asm("ep_32:");
asm("        MOV     R4, R14");
asm("        MOV     R14, PC");
asm("        BX      R12");
asm("        .state32");
asm("        BX      R4");
asm("        .state16");

void delay(int ms)
{
    ms = ms * 1000;

    while (ms--)
        asm(" nop");
}

char to_hex_digit(char n)
{
    return n + (n <= 9 ? '0' : 'A' - 10);
}

// Determine the amount of RAM available, starting from <base> address. The
// base two logarithm is returned. The result, n, fulfills: ram_size >= 2^n
int ram_size(bss_t *bss, volatile char *base)
{
    int offset = 0, power = 0;

    // HACK: Because of a bug in Calypso ROM boot the cmd is always loaded
    // at offset 128k. This bug is not in Calypso Lite but because the ROM
    // boot cannot detect the chip type, this will also be the case when the
    // Lite is used. Because of the less RAM in the Lite chip and the mirror
    // effect, the below ram_size detect does not work
    //bss->chip_id = *((volatile uint16 *) CHIP_ID_CODE);
    if (bss->chip_id == CHIP_ID_CALYPSO_L)
        return 16;  // 2^6 = 64
    if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
        (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A))
        return 16;  // 2^6 = 64

    base[offset] = power;

    // While no RAM mirror effect...
    while (base[offset] == power && base[0] == 0) {
        power++;
        offset = (1 << power);
        base[offset] = power;
    }

    // FIXME: Here we should commence a binary search determining the exact
    // amount of RAM available.

    return power - 1;
}

int is_ram(volatile char *addr)
{
    int i, result = 1;
    char old;

    for (i = 0; i < 2; i++) {
        old = addr[i];
        addr[i] = 0x3A;
        result = (result && addr[i] == 0x3A);
        addr[i] = old;
    }

    return result;
}

int hardware_init(bss_t *bss)
{
    // We read the chipset version directly from the hardware
    bss->chip_id = *((volatile uint16 *) CHIP_ID_CODE);

    // Calypso: Disable internal ROM bootloader
    *(volatile uint16*) 0xfffffb10 |= 0x0300;

    // CSX_MEM_REG format: ----ddd-asswwwww
    //                     7654321076543210
    //   wwwww = wait states
    //      ss = data size: 00=8bit, 01=16bit, 10=32bit
    //       a = write-enable: a=0:ABORT, a=1:write-enable
    //     ddd = dummy cycles on bank switch to faster memory

    *((volatile uint16 *) CS0_MEM_REG) = 0x2A4; // Code memory (Flash)
    *((volatile uint16 *) CS1_MEM_REG) = 0x2A4;
    *((volatile uint16 *) CS2_MEM_REG) = 0x2A4;

    // Peripherals 16bit on C-Sample
    *((volatile uint16 *) CS3_MEM_REG) = 0x2A1;
    // N/A on D-Sample with Calypso in C05 (C035 req'd) 8bit
    *((volatile uint16 *) CS4_MEM_REG) = 0x2A4;
    *((volatile uint16 *) CS6_MEM_REG) = 0x2C0;

    // Reset CLKM register
    //*(volatile uint16 *) MEM_CLKM_ADDR = CLKM_CNTL_ARM_CLK_RST;

    // DSample: The external reset is controlled by the pin
    // "nRESET_OUY/GPIO-7" - this pin stays low after RESET so we have to
    // switch to GPIO mode and set it high.
    *((volatile uint16 *) 0xfffe4806) |= 0x0020; // Enable GPIO module
    *((volatile uint16 *) 0xfffe4804) &= 0xff7f; // Set GPIO-7 = output
    *((volatile uint16 *) 0xfffe4802) |= 0x0080; // Set GPIO-7 = HIGH
    // Switch the pin-mux to GPIO-7 (i.e. release the external RESET)
    *((volatile uint16 *) 0xfffeF00A) |= 0x0008;

    // Raise the external RESET pin CALPLUS JP
    *(volatile unsigned short*) 0xfffffd04 &= ~0x0004;

    // BSample and CSample debug latch
    bss->leds.baddr = (char *)   0x2800000;
    bss->leds.daddr = (uint16 *) 0x2700000;

    led_set(bss, 0);

    // Detect if we have RAM at CS0. Fixed for Calypso+
    if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
        (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A)) {
        if ((bss->ramcs0 = is_ram((char *) 0x04000000)))
            led_on(bss, LED_RAMCS0);
    }
    else
        if ((bss->ramcs0 = is_ram((char *) 0x0)))
            led_on(bss, LED_RAMCS0);

    // Init Flash 'magic' numbers depending on memmap
    if ((bss->chip_id == CHIP_ID_CALYPSO_PLUS) ||
        (bss->chip_id == CHIP_ID_CALYPSO_PLUS_A)) {
        bss->hexAA   = CALP_OFFSET | 0xAA;
        bss->hexAAAA = CALP_OFFSET | 0xAAAA;
        bss->hex5555 = CALP_OFFSET | 0x5555;
        bss->hex0000 = CALP_OFFSET | 0x0000;
        bss->hex0002 = CALP_OFFSET | 0x0002;
    }
    else {
        bss->hexAA   = 0xAA;
        bss->hexAAAA = 0xAAAA;
        bss->hex5555 = 0x5555;
        bss->hex0000 = 0x0000;
        bss->hex0002 = 0x0002;
    }

    return 0;
}

int hardware_init_calypso(bss_t *bss)
{
    char error;

    // Calypso: Set DPLL to 39 MHz (78 MHz/2) and wait for lock
    //*((volatile uint16 *) MEM_DPLL_ADDR) = 0x6332;

    // Calypso: Set DPLL to 52 MHz (104 MHz/2) and wait for lock
    *((volatile uint16 *) MEM_DPLL_ADDR) = 0x6432;

    error = dpll_lock(bss);

    if (!error) {
        // DSample: Note that changing CS4 to be A22 will make the debug latch
        // on the B- and CSamples non-functional.
        *(volatile uint16 *) DBB_CONF_REG |=  0x0008; // Use A22 instead of CS4
    }

    return error;
}

int hardware_init_ulysses(bss_t *bss)
{
    *(volatile uint16 *) DBB_CONF_REG &= ~0x0008; // Use CS4 instead of A22

    return 0;
}

int dpll_lock(bss_t *bss)
{
    int i;

    led_on(bss, LED_BUSY);
    for (i = 0; i < 100000; i++) {
        if (((* (volatile uint16 *) MEM_DPLL_ADDR) & 1) == 0)
            break;
    }
    if (i < 100000) {
        led_off(bss, LED_BUSY);
        i = 0;
    }
    else {
        i = PROTO_ERROR_HARDWARE;
    }
    delay(5);

    return i;
}

void uart_init(bss_t *bss, uint8 my_dlh, uint8 my_dll)
{
    uint8 ch;

    /*
     * Transition from the DSP Rhea bus to the MCU Rhea bus.
     */
    BIT_SET (UART_UIR, UART_MASK_IT);
    BIT_CLR (UART_UIR, UART_ACCESS);
    BIT_CLR (UART_UIR, UART_MASK_IT);

    /*
     * Set the UART in reset mode.
     */
    REGISTER_8_WRITE(bss->uart + UART_MDR1, CMD_UART_RESET);

    /*
     * FIFO configuration
     */
    REGISTER_8_WRITE(bss->uart + UART_LCR, CMD_EFR_EN);
    BIT_SET         (bss->uart + UART_EFR, ENHANCED_EN);
    REGISTER_8_WRITE(bss->uart + UART_LCR, CMD_FCR_MCR_EN);
    REGISTER_8_WRITE(bss->uart + UART_FCR, CMD_FIFO_EN);

    /*
     * Speed configuration
     */
    REGISTER_8_WRITE(bss->uart + UART_LCR, CMD_EFR_EN);
    REGISTER_8_WRITE(bss->uart + UART_DLL, my_dll);
    REGISTER_8_WRITE(bss->uart + UART_DLH, my_dlh);

    /*
     * Character format configuration
     */
    REGISTER_8_WRITE(bss->uart + UART_LCR, CMD_CHAR8_STOP1_NOPAR);

    /*
     * Modem configuration
     */
    REGISTER_8_WRITE(bss->uart + UART_MDR1, CMD_UART_MODE);

    /*
     * Flush Rx buffer
     */
    while ((REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY))
        ch = (REGISTER_8_READ(bss->uart + UART_RHR));
}

/******************************************************************************
 * put/get
 ******************************************************************************/

void putchar(bss_t *bss, uint8 ch)
{
    while (!(REGISTER_8_READ(bss->uart + UART_LSR) & STAT_TXRDY))
        ;
    REGISTER_8_WRITE(bss->uart + UART_THR, ch);
}

void put32(bss_t *bss, unsigned int value)
{
    putchar(bss, value >>  0);
    putchar(bss, value >>  8);
    putchar(bss, value >> 16);
    putchar(bss, value >> 24);
}

void putchar_wait(bss_t *bss)
{
    while (!(REGISTER_8_READ(bss->uart + UART_LSR) & STAT_TX_E))
        ;
}

uint8 getchar_is_ready(bss_t *bss)
{
    return ((REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY));
}

uint8 getchar(bss_t *bss)
{
    led_on(bss, LED_RECV);
    while (!(REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY))
        ;
    led_off(bss, LED_RECV);
    return (uint8) REGISTER_8_READ(bss->uart + UART_RHR);
}

uint8 getchar_fast(bss_t *bss)
{
    while (!(REGISTER_8_READ(bss->uart + UART_LSR) & STAT_RXRDY))
        ;
    return (uint8) REGISTER_8_READ(bss->uart + UART_RHR);
}

uint32 getword(bss_t *bss)
{
    uint32 i;

    i  = ((getchar(bss) & 0xFF)      );
    i |= ((getchar(bss) & 0xFF) <<  8);
    i |= ((getchar(bss) & 0xFF) << 16);
    i |= ((getchar(bss) & 0xFF) << 24);

    return i;
}


/******************************************************************************
 * Flash Functions
 ******************************************************************************/

void flash_detect(bss_t *bss)
{
    uint16 *buf = (uint16 *) bss->buffer;

    if (bss->ramcs0) {
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        *buf++ = 0;
        return;
    }

    ROM_WRITE_ENABLE();

    // Intel devices
    FLASH_WRITE (bss->hex0000, 0xFF);  // Intel read array command
    FLASH_WRITE (bss->hex0000, 0x90);  // Intel intelligent identifier command
    *buf++ = FLASH_READ(bss->hex0000); // Read manufacturer code
    *buf++ = FLASH_READ(bss->hex0002); // Read device code
    FLASH_WRITE (bss->hex0000, 0xFF);  // Intel read array command

    // AMD and SST devices
    FLASH_WRITE (bss->hexAAAA, 0xAA); // AMD and SST unlock cycle 1
    FLASH_WRITE (bss->hex5555, 0x55); // AMD and SST unlock cycle 2
    FLASH_WRITE (bss->hexAAAA, 0x90); // AMD and SST read identifier command
    *buf++ = FLASH_READ(bss->hex0000);
    *buf++ = FLASH_READ(bss->hex0002);
    *buf++ = FLASH_READ(bss->hex0000 + (0xE << 1));  
    *buf++ = FLASH_READ(bss->hex0000 + (0xF << 1));  

    FLASH_WRITE (bss->hexAAAA, 0xAA); // AMD and SST unlock cycle 1
    FLASH_WRITE (bss->hex5555, 0x55); // AMD and SST unlock cycle 2
    FLASH_WRITE (bss->hexAAAA, 0xF0); // AMD and SST reset command

    ROM_WRITE_DISABLE();
}

void flash_query_cfi_data(bss_t *bss, uint16 *data, uint16 offset)
{
    FLASH_WRITE (bss->hexAA,   0xF0); // AMD and SST reset command
    FLASH_WRITE (bss->hex0000, 0xFF); // Intel read-array/reset command
   
    FLASH_WRITE (bss->hexAA,   0x98); // Set CFI query mode

    *data = FLASH_READ(bss->hex0000 | offset);  // Query data

    FLASH_WRITE (bss->hexAA,   0xF0); // AMD and SST reset command
    FLASH_WRITE (bss->hex0000, 0xFF); // Intel read-array/reset command
}


void flash_machine_reset(bss_t *bss)
{
    bss->erase.index = 0;
    bss->erase.list_size = 0;
    bss->erase.addr = NO_ERASE;
}

void led_progress(bss_t *bss, int addr)
{
#if DEBUG_LED_PROGRESS
    int i;

    i = 8 + ((addr >> 13) & 0x7);

    led_clr(bss, 0xFF << 8);
    led_on(bss, i);
#else
#endif
}


/******************************************************************************
 * FIFO Functions
 ******************************************************************************/

void fifo_reset(bss_t *bss)
{
    int x;

    // The fifo buffer size is computed from the RAM available. It is NOT
    // computed as optimally as one could, partly due to the fact that on
    // Calypso devices, the command interpreter is loaded at 128kB offset in
    // RAM and this complicates an exect computation.
    x = 1 << ram_size(bss, (char *) &bss->buffer[0]);
    bss->fifo.max   = 0;
    while (x > sizeof(struct block_s)) {
        x -= sizeof(struct block_s);
        bss->fifo.max++;
    }
    bss->fifo.in    = 0;
    bss->fifo.out   = 0;
    bss->fifo.count = 0;
    bss->fifo.data = (struct block_s *) &bss->buffer[0];
}

int fifo_is_empty(bss_t *bss)
{
    return (bss->fifo.count == 0);
}

int fifo_is_full(bss_t *bss)
{
    return (bss->fifo.count == bss->fifo.max);
}

void fifo_push(bss_t *bss)
{
    int size;
    uint16 *src, *dst;
    int x;

    led_on(bss, LED_OVERHEAD);

    src  = (uint16 *) &bss->rx.size;
    dst  = (uint16 *) &bss->fifo.data[bss->fifo.in];
    size = sizeof(struct block_s) / 2;

    while (size--)
        *dst++ = *src++;

    bss->fifo.count++;
    bss->fifo.in++;
    if (bss->fifo.in >= bss->fifo.max)
        bss->fifo.in = 0;

    if (bss->fifo.count > bss->fifo.max)
        led_on(bss, LED_BUSY);

    led_off(bss, LED_OVERHEAD);
}

struct block_s *fifo_pop(bss_t *bss)
{
    int tmp;

    tmp = bss->fifo.out++;
    bss->fifo.count--;

    if (bss->fifo.out >= bss->fifo.max)
        bss->fifo.out = 0;

    return &bss->fifo.data[tmp];
}

uint32 fifo_peek(bss_t *bss)
{
    return bss->fifo.data[bss->fifo.out].addr;
}


/******************************************************************************
 * Timer Functions
 ******************************************************************************/

void timer_start(bss_t *bss)
{
    bss->time.erase    = 0;
    bss->time.program  = 0;
    bss->time.recvonly = 0;
    bss->time.recv     = 0;
    bss->time.comm     = 0;
    bss->time.setup    = 0;
    bss->time.overhead = 0;
    bss->time.dezip    = 0;

    // Timer load value is maximum.
    *((volatile uint16 *) TIMER1_LOAD) = 0xFFFF;
    bss->time.last     = 0xFFFF;

    // Configure timer with prescaler = 32. Timer step is 1/13MHz * 16 * 32
    // = 39.4us meaning a max range of 2.6s
    *((volatile uint16 *) TIMER1_CNTL) = (4 << 2) | 0x22;

    // Start the timer
    *((volatile uint16 *) TIMER1_CNTL) |= 1;

}

int timer_read(void)
{
    return *((volatile uint16 *) TIMER1_READ);
}

int timer_elapsed_since(bss_t *bss, int last)
{
    int current, elapsed;

    current = *((volatile uint16 *) TIMER1_READ);
    elapsed = last - current;
    if (elapsed < 0)
        elapsed += 0xFFFF;

    return elapsed;
}

int timer_elapsed(bss_t *bss)
{
    int current, elapsed;

    current = *((volatile uint16 *) TIMER1_READ);
    elapsed = bss->time.last - current;
    if (elapsed < 0)
        elapsed += 0xFFFF;

    bss->time.last = current;

    return elapsed;
}


/******************************************************************************
 * Compression
 ******************************************************************************/

#include "../pc/lzdecode.c"


/******************************************************************************
 * Debug
 ******************************************************************************/

#if (DEBUG_TRACE == 1)
#include "debug.c"
#endif
