/******************************************************************************
 * FLUID (Flash Loader Utility Independent of Device)
 *
 * (C) Delta Technologies 2001.
 * Cleanup, modifications and extensions by Mads Meisner-Jensen, mmj@ti.com.
 *
 * Intel Flash Erase and Program Method
 *
 * $Id: intel.c 1.11 Mon, 21 Oct 2002 18:39:13 +0200 mmj $
 *
 ******************************************************************************/

#include "target.h"

//#define DEBUG_PROGRAM
//#define DEBUG_ERASE


static int intel_init(uint32 addr);
static int intel_erase(struct bss_s *bss, uint32 addr);
static int intel_program(struct bss_s *bss, uint16 *dst, uint16 *buf, int size);

const flash_t my_function_table = {
    intel_init,
    intel_erase,
    intel_program,
    0,
    0,
    0,
    0,
    0xAABBCCDD
};


/******************************************************************************
 * Main
 ******************************************************************************/

static int intel_init(uint32 addr)
{
    return PROTO_READY;
}


// Status bits for Intel flash memory devices
#define INTEL_STATE_MACHINE_DONE     (1<<7)
#define INTEL_BLOCK_ERASE_ERROR      (1<<5)
#define INTEL_CMD_SEQ_ERROR          ((1<<4)|(1<<5))
#define INTEL_PROGRAMMING_ERROR      (1<<4)
#define INTEL_VPP_RANGE_ERROR        (1<<3)
#define INTEL_BLOCK_LOCKED_ERROR     (1<<1)


static int intel_erase(struct bss_s *bss, uint32 addr)
{
    int result;
    uint16 status;

#ifdef DEBUG_ERASE
    DEBUG_PUTSTR("E");
    DEBUG_PUTNUM((uint32) addr);
#endif

    FLASH_WRITE(addr, 0x50); // Intel Clear Status Register Command
    FLASH_WRITE(addr, 0xFF); // Intel Read Array Command

    do {
        FLASH_WRITE(addr, 0x60); // Intel Config Setup Command
        FLASH_WRITE(addr, 0xD0); // Intel Unlock Block Command
        FLASH_WRITE((addr + 0x04), 0x90); // Intel Read ID Plane Command
        status = FLASH_READ(addr + 0x04); // Locking changed?
    } while (status & 0x03);
    FLASH_WRITE(addr, 0xFF); // Intel Read Array Command

    FLASH_WRITE(addr, 0x50); // Intel Clear Status Register Command
    FLASH_WRITE(addr, 0x20); // Intel Erase Setup Command
    FLASH_WRITE(addr, 0xD0); // Intel Erase Confirm Command
    
    do {
        if ((result = bss->recv(bss)) != 0)
            return result;
        status = FLASH_READ(addr);
    } while (!(status & INTEL_STATE_MACHINE_DONE));

    status = FLASH_READ(addr);

    if (status & INTEL_VPP_RANGE_ERROR)
        result = PROTO_ERROR_FLASH_VPP;
    else if ((status & INTEL_CMD_SEQ_ERROR) == INTEL_CMD_SEQ_ERROR)
        result = PROTO_ERROR_FLASH_COMMAND;
    else if (status & INTEL_BLOCK_ERASE_ERROR)
        result = PROTO_ERROR_FLASH_TIMEOUT;
    else if (status & INTEL_BLOCK_LOCKED_ERROR)
        result = PROTO_ERROR_FLASH_LOCKED;
    else 
        result = 0;

    FLASH_WRITE(addr, 0xFF); // Intel Read Array Command

#ifdef DEBUG_ERASE
    DEBUG_PUTNUM((uint32) status);
    DEBUG_PUTNUM((uint32) result);
#endif

    return result;
}

static int intel_program(struct bss_s *bss, uint16 *dst, uint16 *buf, int size)
{
    uint16 status;
    uint16 *dst_start = dst;
    int    result = 0;

#ifdef DEBUG_PROGRAM
    DEBUG_PUTSTR("P");
    DEBUG_PUTNUM((uint32) dst);
    DEBUG_PUTNUM((uint32) buf);
    DEBUG_PUTNUM((uint32) size);
#endif

    FLASH_WRITE(dst, 0xFF); // Intel Read Array Command

    FLASH_WRITE(dst, 0x50); // Intel Clear Status Register Command
    FLASH_WRITE(dst, 0xFF); // Intel Read Array Command

    FLASH_WRITE(dst, 0x60); // Intel Config Setup Command
    FLASH_WRITE(dst, 0xD0); // Intel Unlock Block Command
    FLASH_WRITE(dst, 0xFF); // Intel Read Array Command

    while (result == 0 && size--)
    {
        FLASH_WRITE(dst, 0x40); // Intel Program Command
        FLASH_WRITE(dst, *buf);

        // Wait for completion. Reading the flash memory while in program
        // mode allows to get the status.
        do {
            if ((result = bss->recv(bss)) != 0)
                return result;
            status = FLASH_READ(dst);
        } while (!(status & INTEL_STATE_MACHINE_DONE));

        status = FLASH_READ(dst);
        if (status & INTEL_VPP_RANGE_ERROR)
            result = PROTO_ERROR_FLASH_VPP;
        else if (status & INTEL_PROGRAMMING_ERROR)
            result = PROTO_ERROR_FLASH_TIMEOUT;
        else if (status & INTEL_BLOCK_LOCKED_ERROR)
            result = PROTO_ERROR_FLASH_LOCKED;
        else
            result = 0;

        dst++;
        buf++;
    }
    FLASH_WRITE(dst_start, 0xFF); // Intel Read Array Command

    return result;
}

// Assembly code that satisfies compiler's need of an indirect call
asm("        .state16");
asm("IND$CALL:");
asm("        bx      r4");
