/******************************************************************************
 * FLUID (Flash Loader Utility Independent of Device)
 *
 * (C) Delta Technologies 2001.
 * Cleanup, modifications and extensions by Mads Meisner-Jensen, mmj@ti.com.
 *
 * Target common definitions
 *
 * $Id: target.h 1.11.1.12 Mon, 28 Apr 2003 08:49:16 +0200 tsj $
 *
 ******************************************************************************/

#include "protocol.h"
#include "../inc/secure_types.h"

/******************************************************************************
 * Typedefs and Prototypes
 ******************************************************************************/

typedef unsigned char  UBYTE;
typedef unsigned short UINT16;
typedef unsigned long  UINT32;
typedef signed char    INT8;
typedef signed short   INT16;
typedef signed long    INT32;

typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned long  uint32;
typedef signed char    int8;
typedef signed short   int16;
typedef signed long    int32;

#include "../pc/lz.h"

struct bss_s;

typedef union {
    UINT32 i;    // integer value
    UBYTE  b[4]; // byte array
} union32_t;

typedef union {
    UINT16 i;    // integer value
    UBYTE  b[2]; // byte array
} union16_t;

typedef struct {
    int   (*init)(uint32 addr);
    int   (*erase)(struct bss_s *bss, uint32 addr);
    int   (*program)(struct bss_s *bss, uint16 *addr, uint16 *buf, int size);

    uint8 (*getchar)(void);
    void  (*putchar)(uint8 ch);

    void  (*debug_putstr)(uint32 uart, char *string);
    void  (*debug_putnum)(uint32 uart, unsigned int number);

    uint32 uart;
} flash_t;


/******************************************************************************
 * Main Vars
 ******************************************************************************/

#define BUFFER_BLOCK_SIZE (8 * 1024)

// Size (in bytes) of receive buffer for data to be programmed into
// flash. We currently use 8kB of buffer + 1/8 for lz worst-case overhead.
// TODO: Buffer size should be dynamic. +1 for checksum
#define RX_BUFFER_SIZE ((8+1) * 1024)

// Max number of sector addresses in erase list
#define ERASE_LIST_SIZE_MAX (256)

struct block_s {
    uint32 size;
    uint32 addr;
    uint16 data[BUFFER_BLOCK_SIZE / 2];
};

struct bss_s {
    uint32 pc;
    uint32 uart;
    flash_t *flash;
    int  (*recv)(struct bss_s *bss);
    int  recv_state;
    int  lzdec_state;
    int  flash_end;
    int  compress;
    int  ramcs0;
    uint16 chip_id;
    uint16 flash_buf_size; // bytes
    volatile uint32 hexAA;
    volatile uint32 hexAAAA;
    volatile uint32 hex5555;
    volatile uint32 hex0000;
    volatile uint32 hex0002;
    struct led_s {
        volatile char   *baddr; // BSample (and CSample) debug latch addr
        volatile uint16 *daddr; // DSample debug latch addr
        uint16 state;
    } leds;
    struct {
        int erase;
        int program;
        int recvonly;
        int recv;
        int comm;
        int setup;
		int overhead;
		int dezip;
        int start;
        int last;
    } time;
    struct recv_data_s {
        int    index;
        int    rem;
        uint8  cksum;
        // The rest of this structure MUST be the same as block_s
        uint32 size;
        uint32 addr;
        uint8  buf[RX_BUFFER_SIZE];
    } rx;
    struct lz_s lz;
    struct {
        uint32 addr;
        int    index;
        int    list_size;
        uint32 list[ERASE_LIST_SIZE_MAX];
    } erase;
    struct fifo_s {
        struct block_s *data;
        int in;
        int out;
        int max;
        int count;
    } fifo;
    int    buffer_size;
    // IMEI Protection
    T_ROM_SSERVICE         pf_ROM_Sservice;
    T_PLATFORM_CERTIFICATE d_platform_certificate;
    T_SSERVICE_BIND        d_bind;
    // End IMEI Protection
    uint32 buffer[6]; // IMPORTANT: Must be last member of bss!
};
typedef struct bss_s bss_t;


/******************************************************************************
 * Macros
 ******************************************************************************/
#define REGISTER_8_READ(addr)         (*((volatile uint8 *)  (addr)))
#define REGISTER_8_WRITE(addr, data)  (*((volatile uint8 *)  (addr)) = data)
#define REGISTER_16_READ(addr)        (*((volatile uint16 *) (addr)))
#define REGISTER_16_WRITE(addr, data) (*((volatile uint16 *) (addr)) = data)

#define BIT_SET(reg, bit) (*((volatile UINT16 *) (reg)) |=  (1 << (bit)))
#define BIT_CLR(reg, bit) (*((volatile UINT16 *) (reg)) &= ~(1 << (bit)))

#define FLASH_READ(addr)        (*(volatile uint16 *) (addr))
#define FLASH_WRITE(addr, data) (*(volatile uint16 *) (addr)) = data

#define PUTCHAR_FUNCTION_CODE(uart) \
    while (!(REGISTER_8_READ(uart + UART_LSR) & STAT_TXRDY)) ; \
    REGISTER_8_WRITE(uart + UART_THR, ch);

#define GETCHAR_FUNCTION_CODE(uart) \
    while (!(REGISTER_8_READ(uart + UART_LSR) & STAT_RXRDY)) ; \
    return (uint8) REGISTER_8_READ(uart + UART_RHR);

#define GETCHAR_IS_READY_FUNCTION_CODE(uart) \
    return ((REGISTER_8_READ(uart + UART_LSR) & STAT_RXRDY));


/******************************************************************************
 * Debug Macros
 ******************************************************************************/

#define DEBUG_PUTSTR(uart, text) ((flash_t *)(my_addr))->debug_putstr(uart, text);
#define DEBUG_PUTNUM(uart, num)  ((flash_t *)(my_addr))->debug_putnum(uart, num);


/******************************************************************************
 * Hardware Definitions
 ******************************************************************************/

// See CAL207 and CALPLUS207 document for the Calypso (plus) register specification

#define WATCHDOG_TIM_MODE (0xFFFFF804)


#define CHIP_ID_CODE (0xFFFEF000)

#define ARM_ADDRESS (0xFFFFFB00)
#define CS0_MEM_REG (ARM_ADDRESS + 0x00)   /* CS0 memory range */
#define CS1_MEM_REG (ARM_ADDRESS + 0x02)   /* CS1 memory range */
#define CS2_MEM_REG (ARM_ADDRESS + 0x04)   /* CS2 memory range */
#define CS3_MEM_REG (ARM_ADDRESS + 0x06)   /* CS3 memory range */
#define CS4_MEM_REG (ARM_ADDRESS + 0x08)   /* CS4 memory range */
#define CS6_MEM_REG (ARM_ADDRESS + 0x0C)   /* CS6 memory range */

#define CS5_MEM_REG (ARM_ADDRESS + 0x0A)   /* CS5 memory range , CAL+*/

// Pin multiplex config register
#define DBB_CONF_REG (0xFFFEF006) // FIXME: Take name from official software

// ROM/Flash Config register
#define WS_REG_ROM 0x400000   

// UARTs
#define UART_MODEM (0xFFFF5800)
#define UART_IRDA  (0xFFFF5000)

// UART register definitions.
#define UART_RHR   (0x00) /* Receive Holding Register */
#define UART_THR   (0x00) /* Transmit Holding Register */
#define UART_FCR   (0x02) /* FIFO Control Register */ 
#define UART_LCR   (0x03) /* Line Control Register */ 
#define UART_LSR   (0x05) /* Line Status Register */ 
#define UART_MDR1  (0x08) /* Mode Definition Register 1 */
#define UART_EFR   (0x02) /* Enhanced Feature Register */
#define UART_DLL   (0x00) /* Divisor Latches */
#define UART_DLH   (0x01) /* Divisor Latches */
#define UART_UIR   (0xFFFF6000)      /* UART Interface Register */

// UART status bit definitions.
#define STAT_RXRDY (0x01)
#define STAT_RX_OE (0x02)
#define STAT_TXRDY (0x20)
#define STAT_TX_E  (0x40)

// UART_UIR bit definitions.
#define UART_ACCESS  (0)
#define UART_MASK_IT (1)

// UART_EFR bit definitions.
#define ENHANCED_EN  (4)

// UART command bit definitions.
#define CMD_UART_RESET   (0x07)
#define CMD_EFR_EN       (0xBF)
#define CMD_FCR_MCR_EN   (0x80)
#define CMD_FIFO_EN      (0x07)
#define CMD_UART_MODE    (0x00)
#define CMD_CHAR8_STOP1_NOPAR   (0x03)

// Enable/Disable ROM/Flash write strobe
#define ROM_WRITE_ENABLE() *((volatile uint16*) WS_REG_ROM) |= 0x80; 
#define ROM_WRITE_DISABLE() *((volatile uint16*) WS_REG_ROM) &= ~0x80;

// DSample
#define DS_CTRL           0x2780000 // DSample control latch:
#define DS_CTRL_VPP_ENA   0x01 // 12V DC ENAble
#define DS_CTRL_CLK14_ON  0x10 // 14MHz oscillator ON
#define DS_CTRL_CLK14_ENA 0x20 // 14MHz ENAble (13/14MHz MCLK switch)

// ESample
#define CALP_OFFSET 0x4000000  // CS_MODE 0: Flash at 0x04000000 @ CS5

// Secure Calypso Plus
#define CMD_SECURE_CALP_OFFSET 0x08020000 // Flash programmer offset in internal SRAM.
#define MEM_RNG                0x09A00000
#define RNG_ALARM              (MEM_RNG + 0x24)
#define RNG_CONFIG             (MEM_RNG + 0x28)
#define RNG_CONFIG_DEFAULT     0x0834 /* Reset count = 32,
                                       * delay #2 = 6,
                                       * delay #1 = 4 */

// CHIPSET 7

// 
#define MEM_CLKM_ADDR          (0xFFFFFD00)
#define CLKM_CNTL_ARM_CLK      (MEM_CLKM_ADDR + 0x00)
#define CLKM_CNTL_ARM_CLK_RST  (0x1001)

#define CTNL_ARM_CLK           (0xFFFFFD00)      
#define CTNL_CLK               (0xFFFFFD02)      
#define CTNL_RST               (0xFFFFFD04)      

// DPLL register
#define MEM_DPLL_ADDR    (0xffff9800) // DPLL control register
#define DPLL_RESET_VALUE 0x2002       // Reset value of DPLL control register
#define DPLL_LOCKED      0x0001       // Lock status is locked

// Timers
#define TIMER1_CNTL (0xFFFE3800)
#define TIMER1_LOAD (0xFFFE3802)
#define TIMER1_READ (0xFFFE3804)
#define TIMER2_CNTL (0xFFFE6800)
#define TIMER2_LOAD (0xFFFE6802)
#define TIMER2_READ (0xFFFE6804)
